/*
 * manhattan - pu-server.c
 *
 * Copyright (C) 2005..2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/signal.h>
#include <stdarg.h>
#include <syslog.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <compiler.h>
#include <netdb.h>
#include <rtunit-comm.h>
#include <pu-lib.h>
#include <pu-buff.h>
#include <pu-proto.h>
#include <pu-proto-lib.h>

#include <list.h>

/*
 * we only support one listening port and we always want to support
 * IPv4 so we are using IPv6 only when system supports also IPv4
 * on IPv6 sockets.
 */
#if defined(AF_INET6) && defined(IN6_IS_ADDR_V4MAPPED)
#define USE_IPV6	1
#endif

struct c_context {
	int fd;				/* file descriptor */
	struct pu_buff *recv;		/* request receive buffer */
	struct pu_buff *send;		/* currently transmitted skb */

	struct list_head send_queue;
	size_t rcount;			/* number of requests */

	int write_wait;

	int broken;
	int error;

	struct list_head list;
};

static int verbose_level = 1;
static int use_syslog;
static unsigned int rlimit = 16;

static const char *lport = "3000";	/* listen port */
static int max_conns = 20;

static const char *rtunit_device = "/dev/pts/5";
static int rtunit_fd;

static void show_copyright(void)
{
	printf("pu-server " PACKAGE_VERSION ".\n"
"Copyright (C) 2005..2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n");
}

static const char usage[] =
"usage: %s [options]\n"
"Options:\n"
"      --syslog               use syslog\n"
"  -V, --version              show version\n"
"  -c, --conn-limit=LIMIT     connections limit(%d)\n"
"  -f, --file=FILE            shared memory file\n"
"  -h, --help                 show this help\n"
"  -l, --lport=PORT           listen port (3000)\n"
"  -o, --offset=OFFSET        offset in shared memory\n"
"  -q, --quiet                be quiet\n"
"  -s, --shmsize=SIZE         shared memory size\n"
"  -v, --verbose              be verbose\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "pu-server" : name, max_conns);
	exit(EXIT_SUCCESS);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	static struct option long_options[] = {
		{"conn-limit", 1, NULL, 'c'},
		{"file", 1, NULL, 'f'},
		{"help", 0, NULL, 'h'},
		{"lport", 1, NULL, 'l'},
		{"quiet", 0, NULL, 'q'},
		{"syslog", 0, NULL, 1},
		{"verbose", 0, NULL, 'v'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, args, "c:f:hl:n:qvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 1:
			use_syslog = 1;
			break;
		case 'c':
			max_conns = atoi(optarg);
			break;
		case 'f':
			rtunit_device = strdup(optarg);
			break;
		case 'h':
			show_usage(args[0]);
			BUG();
		case 'l':
			lport = strdup(optarg);
			break;
		case 'q':
			verbose_level--;
			break;
		case 'v':
			verbose_level++;
			break;
		case 'V':
			show_version();
			BUG();
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (optind < argc)
		fprintf(stderr, "warning: unknown parameters detected.\n");
}

static int listen_fd = -1;
// static int udp_fd = -1;

static int do_bind(struct addrinfo *ai)
{
	int ret;
	int fd;
	int on = 1;

	fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
	if (fd == -1) {
		log_error("socket: %s\n", strerror(errno));
		return -errno;
	}

	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

	ret = bind(fd, ai->ai_addr, ai->ai_addrlen);
	if (ret < 0) {
		log_error("bind: %s\n", strerror(errno));
		close(fd);
		return -1;
	}
	return fd;
}

static int my_bind(const char *port)
{
	const char *addr = NULL;
	struct addrinfo hints, *ai, *aitop;
	int ret;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	ret = getaddrinfo(addr, port, &hints, &aitop);
	if (ret < 0) {
		fprintf(stderr, "getaddrinfo: %s\n", strerror(errno));
		return -1;
	}

#if USE_IPV6
	for (ai = aitop; ai; ai = ai->ai_next) {
		if (ai->ai_family == AF_INET6) {
			fd = do_bind(ai);
			if (fd >= 0)
				goto out;
		}
	}
#endif
	for (ai = aitop; ai; ai = ai->ai_next) {
		if (ai->ai_family == AF_INET) {
			fd = do_bind(ai);
			if (fd >= 0)
				goto out;
		}
	}
	fd = -1;

out:
	freeaddrinfo(aitop);
	return fd;
}

static int do_listen(void)
{
	int fd;

	if (verbose_level > 1)
		log_info("listening on port %s.\n", lport);
	fd = my_bind(lport);
	if (listen(fd, 5))
		die("listen(): %s.", strerror(errno));
	listen_fd = fd;
	return 0;
}

static int do_udp_connect(struct addrinfo *ai)
{
	int ret;
	int fd;
	int one = 1;

	fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
	if (fd == -1) {
		log_error("socket: %s\n", strerror(errno));
		return -errno;
	}

	ret = setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one));
	if (ret < 0) {
		perror("setsockopt");
		exit(EXIT_FAILURE);
	}

	ret = connect(fd, ai->ai_addr, ai->ai_addrlen);
	if (ret < 0) {
		perror("connect");
		exit(EXIT_FAILURE);
	}
	return fd;
}

static int udp_connect(const char *host, const char *port)
{
	struct addrinfo hints, *ai, *aitop;
	int ret;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE;
	ret = getaddrinfo(host, port, &hints, &aitop);
	if (ret < 0) {
		fprintf(stderr, "getaddrinfo: %s\n", strerror(errno));
		return -1;
	}

#if USE_IPV6
	for (ai = aitop; ai; ai = ai->ai_next) {
		if (ai->ai_family == AF_INET6) {
			fd = do_udp_connect(ai);
			if (fd >= 0)
				goto out;
		}
	}
#endif
	for (ai = aitop; ai; ai = ai->ai_next) {
		if (ai->ai_family == AF_INET) {
			fd = do_udp_connect(ai);
			if (fd >= 0)
				goto out;
		}
	}
	fd = -1;

out:
	freeaddrinfo(aitop);
	return fd;
}

static LIST_HEAD(contexts);
static int conn_count;

static int context_init(struct c_context *c, int fd)
{
	c->fd = fd;
	c->recv = NULL;
	c->send = NULL;
	INIT_LIST_HEAD(&c->send_queue);
	c->rcount = 0;
	c->broken = 0;
	c->write_wait = 0;
	return 0;
}

static int context_destroy(struct c_context *c)
{
	struct pu_buff *skb, *skb2;

	BUG_ON(!conn_count);
	list_for_each_entry_safe(skb, skb2, &c->send_queue, list) {
		list_del(&skb->list);
		pu_free_skb(skb);
	}
	if (c->send)
		pu_free_skb(c->send);
	if (c->recv)
		pu_free_skb(c->recv);
	list_del(&c->list);
	if (verbose_level > 0) {
		if (c->error) {
			log_warn("closing connection: %d: %s\n", c->fd,
					strerror(c->error));
		} else {
			log_warn("gracefully closed connection %d\n", c->fd);
		}
	}
	close(c->fd);
	free(c);
	conn_count--;
	return 0;
}

static int context_write_schedule(struct c_context *c)
{
	/* already scheduled */
	if (c->send != NULL)
		return 0;

	/* nothing to schedule */
	if (list_empty(&c->send_queue))
		return -ENOENT;

	c->send = list_entry(c->send_queue.next, struct pu_buff, list);
	list_del(&c->send->list);
	return 0;
}

static int context_read(struct c_context *c)
{
	struct pu_header *hdr;
	unsigned long len;
	int nowait = 0;

	if (!c->recv)
		c->recv = pu_alloc_skb(512, GFP_KERNEL);

	if (c->broken)
		goto err_close_conn;

	if (c->recv->len < sizeof(*hdr)) {
		c->recv = pu_buff_read(c, c->recv, sizeof(*hdr));
		if (!c->recv)
			goto err_close_conn;
		if (c->broken)
			c->write_wait = 1;
		if (c->recv->len < sizeof(*hdr))
			return 0;
		nowait = 1;
	}

	hdr = (void *) c->recv->data;
	if (hdr->magic != PU_MAGIC) {
		log_error("protocol error: invalid magic %02x\n", hdr->magic);
		goto err_close_conn;
	}

	/*
	 * we don't care about packets with error code
	 */
	len = be32_to_cpu(hdr->data_bytes);
	if (len >= 65536) {
		log_error("protocol error: too large message %ld\n", len);
		goto err_close_conn;
	}
	len += sizeof(*hdr);

	if (pu_align_data(len) - c->recv->len > 0) {
		if (nowait)
			return 0;

		c->recv = pu_buff_read(c, c->recv,
				pu_align_data(len) - c->recv->len);
		if (!c->recv)
			goto err_close_conn;
		if (c->broken) {
			c->write_wait = 1;
			return 0;
		}
	}

	if (c->recv->len < pu_align_data(len))
		return 0;

	pu_skb_trim(c->recv, len);

	c->rcount++;

	switch (hdr->proto) {
	case PU_PROTO_GRADER:
		if (pu_proto_grader(c, c->recv) < 0)
			goto err_close_conn;
//			c->broken = 1;
		break;
	default:
		log_error("protocol warning: unsupported protocol 0x%02x\n",
				hdr->proto);
		pu_proto_error(c, c->recv, PU_EPROTONOSUPPORT);
		break;
	}
	c->recv = NULL;
	return 0;

err_close_conn:
	context_destroy(c);
	return -ECONNRESET;
}

/*
 * send response to client
 */
static int context_write(struct c_context *c)
{
	struct pu_buff *skb;
	int ret = 0;

	if (c->send == NULL)
		context_write_schedule(c);
	skb = c->send;

	if (skb->len) {
		ret = write(c->fd, skb->data, skb->len);
		if (ret < 0)
			return -errno;
		if (!ret)
			return -ECONNRESET;
		pu_skb_pull(skb, ret);
	}

	if (!skb->len) {
		c->rcount--;
		pu_consume_skb(skb);
		c->send = NULL;
		context_write_schedule(c);
	}
	return 0;
}

static void accept_connection(void)
{
	int ret;
	struct c_context *c;
	int op;

	ret = accept(listen_fd, NULL, NULL);
	if (ret < 0) {
		fprintf(stderr, "accept error: %s", strerror(errno));
		return;
	}
	if (conn_count >= max_conns) {
		fprintf(stderr, "accept: connections limit exceed.\n");
		close(ret);
		return;
	}

	if (verbose_level > 0)
		fprintf(stderr, "accepted connection(%d): fd = %d\n",
				conn_count, ret);

	if (fcntl(ret, F_SETFL, O_NONBLOCK)) {
		fprintf(stderr, "fcntl(): %s", strerror(errno));
		close(ret);
		return;
	}

	op = 1;
	setsockopt(ret, SOL_TCP, TCP_NODELAY, &op, sizeof(op));
	setsockopt(ret, SOL_SOCKET, SO_KEEPALIVE, &op, sizeof(op));

	c = xmalloc(sizeof(*c));
	context_init(c, ret);
	list_add_tail(&c->list, &contexts);
	conn_count++;
}

int pu_proto_xmit(void *priv, struct pu_buff *skb)
{
	struct c_context *c = priv;

	list_add(&skb->list, &c->send_queue);
	if (c->send == NULL)
		context_write_schedule(c);
	return 0;
}

struct pu_buff *pu_buff_read(struct c_context *c, struct pu_buff *skb,
		size_t count)
{
	unsigned char *p;
	ssize_t ret;

	if (pu_skb_tailroom(skb) < count) {
		struct pu_buff *skb2 = skb_copy_expand(skb, 0, count,
				GFP_KERNEL);
		pu_consume_skb(skb);
		if (skb == NULL)
			return skb;
		skb = skb2;
	}
	p = pu_skb_put(skb, 0);

	ret = xread(c->fd, p, count);
	if (ret > 0) {
		pu_skb_put(skb, ret);
	} else {
		c->broken = 1;
		c->error = -ret;
	}
	return skb;
}

struct pevent_msg {
	be32 type;
	be32 position;
	be32 events;
};

#define PU_PROTO_GRADER	2

static int pu_proto_xmit_new_fd(int fd, struct pu_buff *skb,
		unsigned int proto, unsigned int id)
{
	struct pu_header *hdr;
	void *p;
	unsigned int padding;
	unsigned int len;

	len = skb->len;

	padding = pu_align_data(len) - len;
	if (padding) {
		p = pu_skb_put(skb, padding);
		memset(p, 0, padding);
	}

	hdr = (void *)pu_skb_push(skb, sizeof(*hdr));
	hdr->magic = PU_MAGIC;
	hdr->proto = proto;
	hdr->id = cpu_to_be16(id);
	hdr->data_bytes = cpu_to_be32(len);
//	pu_proto_hdr_init(hdr, id, proto, len);
	write_or_die(fd, skb->data, skb->len);
	pu_consume_skb(skb);
	return 0;
}

void pevent_callback(unsigned int pos, unsigned int events)
{
	struct pu_buff *skb;
	struct pevent_msg *data;

	skb = pu_proto_alloc_skb(sizeof(*data));
	data = (void *) pu_skb_put(skb, sizeof(*data));
	data->type = cpu_to_be32(2);
	data->position = cpu_to_be32(pos);
	data->events = cpu_to_be32(events);
//	pu_proto_xmit_new_fd(udp_fd, skb, PU_PROTO_GRADER, 0);
//	write(udp_fd, &pos, sizeof(pos));
}

static int main_loop(void)
{
	struct pollfd *pfd;
	int ret;
	int cfd;
	struct c_context *c;
	struct list_head *t;

	pfd = xmalloc((max_conns + 3) * sizeof(struct pollfd));
	for (;;) {
		int timeout;

		/* prepare for poll */
		pfd[0].fd = listen_fd;
		pfd[0].events = POLLIN;

		pfd[1].fd = rtunit_fd;
		pfd[1].events = POLLIN;
		timeout = rtunit_comm_timeout();
		if (rtunit_comm_want_out())
			pfd[1].events |= POLLOUT;
		cfd = 2;

		list_for_each_entry(c, &contexts, list) {
			pfd[cfd].fd = c->fd;
			pfd[cfd].events = 0;
			if (c->rcount < rlimit)
				pfd[cfd].events |= POLLIN;
			if (c->send)
				pfd[cfd].events |= POLLOUT;
			cfd++;
		}

		ret = poll(pfd, cfd, timeout);
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			die_errno("poll() error:");
		} else if (!ret)
			continue;

		cfd = 2;
		for (t = list_next(&contexts); t != &contexts;) {
			c = list_entry(t, struct c_context, list);
			t = list_next(t);

			BUG_ON(c->fd != pfd[cfd].fd);
			if (pfd[cfd].revents & (POLLERR | POLLHUP | POLLNVAL)) {
				context_destroy(c);
			} else if (c->write_wait && !c->rcount) {
				context_destroy(c);
			} else {
				ret = 0;
				if (pfd[cfd].revents & pfd[cfd].events & POLLIN)
					ret = context_read(c);

				if ((pfd[cfd].revents & POLLOUT)
						&& (ret != -ECONNRESET)) {
					ret = context_write(c);
					if (ret == -ECONNRESET)
						context_destroy(c);
				}
			}
			cfd++;
		}

		if (pfd[0].revents & POLLIN)
			accept_connection();

		if (pfd[1].revents & POLLIN)
			rtunit_comm_pollin(pfd[1].fd);
		if (pfd[1].revents & POLLOUT)
			rtunit_comm_pollout(pfd[1].fd);
	}
	return 0;
}

int main(int argc, char **args)
{
	signal(SIGPIPE, SIG_IGN);
	parse_command_line(argc, args);

	logger_init();
	rtunit_fd = rtunit_comm_init(rtunit_device);
	if (rtunit_fd < 0)
		die("cannot initialize rtunit communication");
	do_listen();
//	udp_fd = udp_connect("192.168.3.255", "5000");
	main_loop();
	return 0;
}
